@@ -221,6 +221,9 @@ gok_scanner_find_node (xmlNode *pNode, const xmlChar* name, const xmlChar* targe
 	int      target_lang_len = strlen ((char *)target_lang);
 	gboolean lang_is_matched = FALSE;	
 
+	gok_log_enter();
+	gboolean is_lang_match = FALSE;
+	gboolean is_perfect_match = FALSE;
 	/* loop through the document looking for the name string */
 	while (pNode != NULL)
 	{
@@ -228,8 +231,6 @@ gok_scanner_find_node (xmlNode *pNode, const xmlChar* name, const xmlChar* targe
 		if (xmlStrcmp (pNode->name, name) == 0)
 		{
 			xmlChar *lang = NULL;
-			gboolean is_lang_match = FALSE;
-			gboolean is_perfect_match = FALSE;
 			/* If xml:lang matches our locale, return, else keep looking... */
 			lang = xmlNodeGetLang (pNode);
 			if (lang) {
@@ -240,6 +241,8 @@ gok_scanner_find_node (xmlNode *pNode, const xmlChar* name, const xmlChar* targe
 				retval = pNode;
 				if (!xmlStrncmp (lang, target_lang, target_lang_len)) { /* perfect match */
 					if (lang && node_lang) *node_lang = lang;
+					gok_log ("found node");
+					gok_log_leave();
 					return retval;
 				}
 				else if (is_lang_match) {
@@ -257,6 +260,7 @@ gok_scanner_find_node (xmlNode *pNode, const xmlChar* name, const xmlChar* targe
 		pNode = pNode->next;
 	}
 
+	gok_log_leave();
 	return retval;
 }
 
@@ -524,10 +528,10 @@ gboolean gok_scanner_read_access_method (gchar* Filename)
 	strcpy (pAccessMethod->DisplayName, (gchar *)pStringAttributeValue);
 
 	/* read the UI for the access method */
-	gok_scanner_read_description (pDoc, pNodeRoot->xmlChildrenNode, pAccessMethod);
-	gok_scanner_read_operation (pDoc, pNodeRoot->xmlChildrenNode, pAccessMethod);
-	gok_scanner_read_feedback (pDoc, pNodeRoot->xmlChildrenNode, pAccessMethod);
-	gok_scanner_read_options (pDoc, pNodeRoot->xmlChildrenNode, pAccessMethod);
+	gok_scanner_read_description (pDoc, pNodeAccessmethod->xmlChildrenNode, pAccessMethod);
+	gok_scanner_read_operation (pDoc, pNodeAccessmethod->xmlChildrenNode, pAccessMethod);
+	gok_scanner_read_feedback (pDoc, pNodeAccessmethod->xmlChildrenNode, pAccessMethod);
+	gok_scanner_read_options (pDoc, pNodeAccessmethod->xmlChildrenNode, pAccessMethod);
 
 	/* read the rates for this effect */
 	gok_scanner_read_rates (pNodeRoot->xmlChildrenNode, pAccessMethod);
@@ -973,6 +977,7 @@ gboolean gok_scanner_read_description (xmlDoc* pDoc, xmlNode* pNode, GokAccessMe
 		return TRUE;
 	}
 	else {
+		strncpy (pAccessMethod->Description, (gchar *)"error reading description", MAX_DESCRIPTION_TEXT);
 		gok_log_leave();
 		return FALSE;
 	}
