@@ -40,15 +40,14 @@
 #include "gok-modifier.h"
 #include "gok-data.h"
 #include "gok-modifier-keymasks.h"
+#include "gok-gconf-keys.h"
 
 /*
- * these definitions allow us to tweak the efficiency/accuracy of the
- * gui search
+ * NOTE: we can tweak the efficiency/accuracy of the
+ * gui search by adjusting using the spy/gui_search_depth and 
+ * spy/gui_search_breadth gconf keys.
  */
 
-#define GOKSPY_GUI_BREADTH 999
-#define GOKSPY_GUI_DEPTH 999
-
 /*
  * NOTES
  *
@@ -62,8 +61,6 @@
 /*
  * private prototypes
  */
-static gboolean gok_spy_is_menu_role (AccessibleRole role);
-
 static void gok_spy_focus_listener (const AccessibleEvent *event,
                                     void *user_data);
 
@@ -97,13 +94,16 @@ static AccessibleNode*  gok_spy_append_node (AccessibleNode* pnode,
 static void gok_spy_modifier_listener(const AccessibleEvent *event,
                                       void *user_data);
 
-static gboolean gok_spy_worth_searching (Accessible* accessible);
+static gboolean gok_spy_worth_searching (Accessible* accessible, AccessibleRole role);
 
 static void gok_spy_window_activate_listener (const AccessibleEvent *event,
                                               void *user_data);
 
 static void gok_spy_window_deactivate_listener (const AccessibleEvent *event,
                                                 void *user_data);
+												
+static gboolean gok_spy_is_visible (Accessible *accessible);
+												
 
 /*
  * gok-spy variables
@@ -131,6 +131,9 @@ static gint explicitrefs;
 static gint explicitunrefs;
 static gint implicitrefs;
 static gulong keyboardmods;
+static gboolean spyshutdown = FALSE;/* this global could be removed if we have a
+						way of removing our idle handler in gok_spy_stop */
+
 
 /* 
  * for debugging
@@ -145,6 +148,7 @@ void gok_spy_accessible_ref(Accessible* accessible)
 	{
 		explicitrefs++;
 		Accessible_ref (accessible);
+#if defined(ENABLE_LOGGING_NORMAL)
 		s = Accessible_getName (accessible);
 		if (s == NULL)
 		{
@@ -155,6 +159,7 @@ void gok_spy_accessible_ref(Accessible* accessible)
 			gok_log ("%#x %s", accessible, s);
 			SPI_freeString (s);
 		}
+#endif
 	}
 	else
 	{
@@ -200,6 +205,7 @@ void gok_spy_accessible_unref(Accessible* accessible)
 
 	if (accessible != NULL)
 	{
+#if defined(ENABLE_LOGGING_NORMAL)
 		s = Accessible_getName (accessible);
 		if (s == NULL)
 		{
@@ -210,6 +216,7 @@ void gok_spy_accessible_unref(Accessible* accessible)
 			gok_log ("%#x %s", accessible, s);
 			SPI_freeString (s);
 		}
+#endif
 		explicitunrefs++;
 		Accessible_unref (accessible);
 	}
@@ -252,6 +259,13 @@ gboolean gok_spy_accessible_is_okay(Accessible* accessible)
 	return returncode;
 }
 
+static gboolean
+gok_spy_idle_handler (gpointer data)
+{
+	return gok_spy_check_queues ();
+}
+
+
 /**
  * gok_spy_open:
  *
@@ -317,6 +331,62 @@ void gok_spy_open()
 	gok_log_leave ();
 }
 
+/** 
+ * gok_spy_stop:
+ *
+ * Call this to unhook all the gok_spy internal listeners. Do this as as the
+ * first thing when shutting down the application.
+ */
+void gok_spy_stop (void)
+{
+	gboolean result = FALSE;	
+	
+	spyshutdown = TRUE;
+	
+	if (focusListener != NULL) 
+	{
+		result = SPI_deregisterGlobalEventListenerAll(focusListener);
+		gok_log("deregistering focus listener returned: %d",result);
+		AccessibleEventListener_unref(focusListener);
+		focusListener = NULL;
+	}
+	if (defunctnessListener != NULL) 
+	{
+		result = SPI_deregisterGlobalEventListenerAll(defunctnessListener);
+		gok_log("deregistering defunctness listener returned: %d",result);
+		AccessibleEventListener_unref(defunctnessListener);
+		defunctnessListener = NULL;
+	}
+	if (mouseListener != NULL)
+	{
+		result = SPI_deregisterDeviceEventListener (mouseListener, NULL);
+		gok_log("deregistering mouse listener returned: %d",result);
+		AccessibleDeviceListener_unref(mouseListener);
+		mouseListener = NULL;
+	}
+	if (modifierListener != NULL)
+	{
+		result = SPI_deregisterGlobalEventListenerAll (modifierListener);
+		gok_log("deregistering modifier listener returned: %d",	result);
+		AccessibleEventListener_unref(modifierListener);
+		modifierListener = NULL;
+	}
+	if (windowActivateListener != NULL)
+	{
+		result = SPI_deregisterGlobalEventListenerAll (windowActivateListener);
+		gok_log("deregistering window activate listener returned: %d",result);
+		AccessibleDeviceListener_unref(windowActivateListener);
+		windowActivateListener = NULL;
+	}
+	if (windowDeactivateListener != NULL)
+	{
+		result= SPI_deregisterGlobalEventListenerAll (windowDeactivateListener);
+		gok_log("deregistering window deactivate listener returned: %d",result);
+		AccessibleEventListener_unref(windowDeactivateListener );
+		windowDeactivateListener = NULL;
+	}
+}
+	
 /** 
  * gok_spy_close:
  *
@@ -325,7 +395,6 @@ void gok_spy_open()
 void gok_spy_close(void)
 {
 	EventNode* en;
-	gboolean result = FALSE;	
 	
 	gok_log_enter ();
 
@@ -333,34 +402,8 @@ void gok_spy_close(void)
 	{
 		m_gokSpyOpen = FALSE;
 		
-		if (focusListener != NULL) 
-		{
-			result = SPI_deregisterGlobalEventListenerAll(focusListener);
-			gok_log("deregistering focus listener returned: %d",result);
-		}
-		if (defunctnessListener != NULL) 
-		{
-			result = SPI_deregisterGlobalEventListenerAll(defunctnessListener);
-			gok_log("deregistering defunctness listener returned: %d",result);
-		}
-		if (mouseListener != NULL)
-		{
-			result = SPI_deregisterDeviceEventListener (mouseListener, NULL);
-			gok_log("deregistering mouse listener returned: %d",result);
-		}
-		if (modifierListener != NULL)
-		{
-			result = SPI_deregisterGlobalEventListenerAll(
-				modifierListener);
-			gok_log("deregistering modifier listener returned: %d",
-				result);
-		}
-	
-		AccessibleEventListener_unref(focusListener);
-		AccessibleEventListener_unref(defunctnessListener);
-		AccessibleDeviceListener_unref(mouseListener);
-		AccessibleEventListener_unref(modifierListener);
-
+		gok_spy_stop ();
+		
 		if (m_ptheWindowAccessible != NULL) gok_spy_accessible_unref(m_ptheWindowAccessible);
 		if (m_ptheAccessibleWithText != NULL) gok_spy_accessible_unref(m_ptheAccessibleWithText);
 
@@ -539,8 +582,13 @@ gboolean gok_spy_has_child (Accessible* accessible, GokSpySearchType type, Acces
 	gok_log_enter();
 	if (gok_spy_accessible_is_okay(accessible) == TRUE)
 	{
-		gok_spy_has_child_aux (accessible, type, role, 
-				       GOKSPY_GUI_BREADTH, GOKSPY_GUI_DEPTH, 0, &bfound);
+		gint breadth;
+		gint depth;
+		GConfClient* client = gok_data_get_gconf_client ();
+		gok_gconf_get_int ( client, GOK_GCONF_SPY_SEARCH_BREADTH, &breadth );
+		gok_gconf_get_int ( client, GOK_GCONF_SPY_SEARCH_DEPTH, &depth );
+		gok_spy_has_child_aux (accessible, type, role, breadth, depth, 
+			0, &bfound);
 	}
 	else
 	{
@@ -562,12 +610,18 @@ gboolean gok_spy_check_queues(void)
 	char *s;
 
 	gok_log_enter();
+	
+	if (spyshutdown) {
+		return FALSE;
+	}
+	
 	while (eventQueue && g_queue_is_empty(eventQueue) == FALSE)
 	{
 		en = (EventNode*)g_queue_pop_head(eventQueue);
 
 		gok_log ("eventQueue->length = %d", eventQueue->length);
 
+#if defined(ENABLE_LOGGING_NORMAL)
 		if (en->paccessible != NULL)
 		{
 			s = Accessible_getName (en->paccessible);
@@ -581,6 +635,7 @@ gboolean gok_spy_check_queues(void)
 				SPI_freeString (s);
 			}
 		}
+#endif
 
 		switch (en->type)
 		{
@@ -602,9 +657,10 @@ gboolean gok_spy_check_queues(void)
 			gok_log("GOKSPY_DEFUNCT_EVENT");
 			/* note: currently only defunct events for the current application 
 				are pushed on the queue */
-			gok_spy_accessible_unref (m_ptheWindowAccessible); /* for our 
-				reference book-keeping */
+			gok_spy_accessible_unref (m_ptheWindowAccessible);
+			gok_spy_accessible_unref (m_ptheAccessibleWithText);
 			m_ptheWindowAccessible = NULL;
+			m_ptheAccessibleWithText = NULL;
 			((AppChangeListenerCB)(m_ptheWindowChangeListener))(m_ptheWindowAccessible);
 			break;
 		case GOKSPY_WINDOW_DEACTIVATE_EVENT:
@@ -620,7 +676,9 @@ gboolean gok_spy_check_queues(void)
 			 */
 			if (en->paccessible == m_ptheWindowAccessible) {
 				gok_spy_accessible_unref (m_ptheWindowAccessible);
+				gok_spy_accessible_unref (m_ptheAccessibleWithText);
 				m_ptheWindowAccessible = NULL;
+				m_ptheAccessibleWithText = NULL;
 				((AppChangeListenerCB)(m_ptheWindowChangeListener))(m_ptheWindowAccessible);
 			}
 			break;
@@ -646,11 +704,13 @@ gboolean gok_spy_check_queues(void)
 */ 
 gboolean gok_spy_check_window(AccessibleRole role)
 {
+	/* TODO - improve efficiency here? */
 	if ((role == SPI_ROLE_WINDOW) ||
 	    (role == SPI_ROLE_DIALOG) ||
 	    (role == SPI_ROLE_FILE_CHOOSER) ||
 	    (role == SPI_ROLE_FRAME) ||	
-	    (role == SPI_ROLE_DESKTOP_FRAME))
+	    (role == SPI_ROLE_DESKTOP_FRAME) ||
+		(role == SPI_ROLE_FONT_CHOOSER))
 	{
 		return TRUE;
 	}
@@ -658,28 +718,63 @@ gboolean gok_spy_check_window(AccessibleRole role)
 }
 
 /** 
-* gok_spy_worth_searching:
-*
-* @accessible: the accessible (possibly a parent) to examine for worthiness.
-*
-* This function decides if the accessible might have a subtree worth searching.
-*
-* Returns: boolean success. 
+ * gok_spy_worth_searching:
+ *
+ * @accessible: the accessible (possibly a parent) to examine for worthiness.
+ * @role: the role we are searching for.
+ *
+ * This function decides if the accessible might have a subtree worth searching.
+ *
+ * Returns: boolean success. 
 */ 
-gboolean gok_spy_worth_searching (Accessible* accessible)
+gboolean gok_spy_worth_searching (Accessible* accessible, AccessibleRole role)
 {
 	gboolean worthiness;
+	gboolean bmenu;
 	AccessibleStateSet* ass = NULL;
-	worthiness = TRUE;
+
 	gok_log_enter();
+	bmenu = gok_spy_is_menu_role (role);
+	if (bmenu)
+	{		
+		switch (Accessible_getRole(accessible))
+		{
+			/* using this breaks mozilla menus:
+			case SPI_ROLE_TOOL_BAR: */
+			case SPI_ROLE_STATUS_BAR:
+			case SPI_ROLE_COMBO_BOX:
+				/* add cases here */
+				gok_log("aborting menu search");
+				gok_log_leave();
+				return FALSE;
+		}
+	}
+	else
+	{
+		switch (Accessible_getRole(accessible))
+		{
+			case SPI_ROLE_MENU_BAR:
+			case SPI_ROLE_MENU:
+			case SPI_ROLE_COMBO_BOX:
+				/* add cases here */
+				gok_log("aborting search");
+				gok_log_leave();
+				return FALSE;
+		}
+	}
+	
+	worthiness = TRUE;
 	ass = Accessible_getStateSet(accessible);
 	if (ass != NULL)
 	{
 		AccessibleRole role = Accessible_getRole (accessible);
-		/* assumption: if this thing manages its own children, then it 
+		/* state heuristic: 
+		 * - if this thing manages its own children, then it 
 		 * is probably not worth traversing.
+		 * - if it is a menu-like thing then it doesn't need to be "showing"
 		 */
-		if ( AccessibleStateSet_contains( ass, SPI_STATE_MANAGES_DESCENDANTS ))
+		if (AccessibleStateSet_contains( ass, SPI_STATE_MANAGES_DESCENDANTS ) ||
+			(!bmenu && !AccessibleStateSet_contains( ass, SPI_STATE_SHOWING )) )
 		{
 			worthiness = FALSE;
 		}
@@ -715,11 +810,12 @@ gboolean gok_spy_has_child_aux (Accessible* accessible,
     AccessibleStateSet *states;
     int child_count;
     int i; 
+	gboolean nameless_menu = FALSE;
 
     if (accessible == NULL) return *found; 
 	if (cdepth++ == depth) return *found;
 	
-	if (gok_spy_worth_searching(accessible) == FALSE) return *found;
+	if (gok_spy_worth_searching(accessible, role) == FALSE) return *found;
 	
     child_count = Accessible_getChildCount (accessible);
     
@@ -743,24 +839,28 @@ gboolean gok_spy_has_child_aux (Accessible* accessible,
 		    {
 			    if (Accessible_getRole (pAccessibleChild) == role) {
 				    gok_log ("gok_spy_has_child_aux: found child at depth %d", cdepth);
-				    gok_spy_accessible_unref(pAccessibleChild);
-				    *found = TRUE;
-				    return *found;
+					*found = TRUE;
+					gok_spy_accessible_unref(pAccessibleChild);
+					return *found;
 			    }
 		    }
 		    else if (type == GOK_SPY_SEARCH_MENU) {
 			    AccessibleRole a_role = Accessible_getRole (pAccessibleChild);
 			    if ((a_role == role) || 
-				gok_spy_is_menu_role (a_role)) {
+				 (a_role == SPI_ROLE_MENU)) {
 				    gchar *name = Accessible_getName (pAccessibleChild);
 				    /* menu items must have either text or an image! */
 				    if (name && (strlen (name) > 0)) { /* todo: what about icon-menus? */
-					    gok_log ("gok_spy_has_child_aux: found child at depth %d", cdepth);
+					    gok_log ("gok_spy_has_child_aux: found [%s] at depth %d", name, cdepth);
 					    gok_spy_accessible_unref(pAccessibleChild);
 					    SPI_freeString (name);
 					    *found = TRUE;					    
 					    return *found;
 				    }
+					else {
+						nameless_menu = TRUE;
+					}
+						
 			    }
 		    }
 		    else if (type == GOK_SPY_SEARCH_EDITABLE_TEXT) {
@@ -770,7 +870,11 @@ gboolean gok_spy_has_child_aux (Accessible* accessible,
 				    return *found;
 			    }
 		    }
-		    gok_spy_has_child_aux (pAccessibleChild, type, role, children, depth, cdepth, found);
+			/* ignore children of a nameless menu, and abort if we have already
+			   found one target */
+			if (!*found && !nameless_menu )
+		    	gok_spy_has_child_aux (pAccessibleChild, type, role, 
+					children, depth, cdepth, found);
 		    gok_spy_accessible_unref(pAccessibleChild);
 		} 
 	}
@@ -792,7 +896,7 @@ gok_spy_is_visible (Accessible *accessible)
 	return retval;
 }
 
-static gboolean
+gboolean
 gok_spy_is_menu_role (AccessibleRole role)
 {
 	return ((role  ==  SPI_ROLE_MENU_ITEM) ||
@@ -855,7 +959,7 @@ void gok_spy_find_children (AccessibleNode* pNode, GokSpySearchType type, Access
 
 	pAccessibleParent = pNode->paccessible; 
 
-	if (gok_spy_worth_searching(pAccessibleParent) == FALSE) return;
+	if (gok_spy_worth_searching(pAccessibleParent, role) == FALSE) return;
 
 	child_count = Accessible_getChildCount (pAccessibleParent);
 
@@ -1036,18 +1140,21 @@ AccessibleNode* gok_spy_append_node (AccessibleNode* pNode,
 		AccessibleHypertext_unref (hypertext);
 	}
 
-	/* if the name is empty then try LABELLED_BY, else ignore this child */ 
+	/* if the name is empty then try LABELLED_BY, else assign "???" */ 
 	else {
 		/* look for a label */
 		if (pName == NULL)
 		{
 			gok_log("no name, so looking at relations...");
 			relations = Accessible_getRelationSet(pAccessible);
+			/*
 			if (relations == NULL)	{
 				gok_log("no relationset for this object");
 				gok_log_leave();
 				return pNewNode;
 			}
+			*/
+			if (relations != NULL)
 			while (relations[i]) {
 				type = AccessibleRelation_getRelationType(relations[i]);
 				if (type == SPI_RELATION_LABELED_BY) {
@@ -1072,6 +1179,7 @@ AccessibleNode* gok_spy_append_node (AccessibleNode* pNode,
 					}
 					break; /* from while loop */
 				}
+				i++;
 			}
 		}
 		
@@ -1080,6 +1188,13 @@ AccessibleNode* gok_spy_append_node (AccessibleNode* pNode,
 			/* create a new node */
 			gok_spy_add_node (pNode, pAccessible, -1, pName);
 		}
+		/* if we decide to allow access to nameless menus ...
+		else if ( Accessible_getRole (pAccessible) == SPI_ROLE_MENU ) {
+			pName = g_strdup (Accessible_getRoleName(pAccessible));
+			gok_log ("Can't think of a good name so using role:[%s]",pName);
+			gok_spy_add_node (pNode, pAccessible, -1, pName);
+		}
+		*/
 		else if (Accessible_isText (pAccessible)) {
 			/* 
 			* sadly, many text objects are still nameless - yet we want 
@@ -1135,8 +1250,10 @@ AccessibleNode* gok_spy_append_node (AccessibleNode* pNode,
 				gok_spy_add_node (pNode, pAccessible, -1, pName);
 			}
 		}
+			
 	}
 	
+	
 	while (pNode->pnext) 
 		pNode = pNode->pnext;
 
@@ -1181,7 +1298,7 @@ void gok_spy_defunctness_listener (const AccessibleEvent *event, void *user_data
 		en->type = GOKSPY_DEFUNCT_EVENT;
 		g_queue_push_tail ( eventQueue, en );
 		gok_spy_add_idle_handler ();
-		gok_log_x("our window is gone!");
+		gok_log ("Our target window is now defunct!");
 	}
 	
 	gok_log_leave();
@@ -1426,12 +1543,6 @@ void gok_spy_modifier_listener(const AccessibleEvent *event, void *user_data)
 	gok_log_leave();
 }
 
-static gboolean
-gok_spy_idle_handler (gpointer data)
-{
-	return gok_spy_check_queues ();
-}
-
 /** 
  * gok_spy_add_idle_handler:
  *
@@ -1442,6 +1553,75 @@ void gok_spy_add_idle_handler (void)
 	g_idle_add (gok_spy_idle_handler, NULL);
 }
 
+/* 
+* gok_spy_get_active_frame
+* 
+* User must call gok_spy_free when finished with this list.
+* 
+* Returns: pointer to the active Accessible* or NULL
+*/
+Accessible* 
+gok_spy_get_active_frame( )
+{
+	gint i;
+	gint j;
+	gint k;
+	gint n_desktops;
+	gint n_apps;
+	gint n_frames;
+	Accessible* desktop;
+	Accessible* child;
+	AccessibleStateSet* ass = NULL;
+
+	gok_log_enter();
+
+	n_desktops = SPI_getDesktopCount();
+	for (i = 0; i < n_desktops; i++)
+	{
+		desktop = SPI_getDesktop(i);
+		n_apps= Accessible_getChildCount(desktop);
+		for (j = 0; j < n_apps; j++)
+		{
+			child = Accessible_getChildAtIndex(desktop, j);
+			
+			if (!child)
+				continue;
+			
+			/* Applications are not active
+			   so we must look for a child frame */
+			n_frames = Accessible_getChildCount (child);
+			for (k = 0; k < n_frames; k++) {
+				Accessible *frame;		
+				frame = Accessible_getChildAtIndex (child, k);
+
+				if (!frame)
+					continue;
+
+				ass = Accessible_getStateSet (frame); /* implicit ref? */
+				
+				if (!ass) {
+					Accessible_unref (frame);
+					continue;
+				}
+
+				if (AccessibleStateSet_contains (ass, SPI_STATE_ACTIVE)) {
+					gok_log ("Found active frame");
+					Accessible_unref (desktop);
+					AccessibleStateSet_unref (ass);
+					Accessible_unref (child);
+					return frame;
+				}
+				AccessibleStateSet_unref (ass);
+				Accessible_unref (frame);
+			}
+			Accessible_unref (child);
+		}
+		Accessible_unref (desktop);
+	}
+	gok_log_leave();
+	return NULL;
+}
+
 /*
 gboolean gok_spy_is_desktop(Accessible* pAccessible)
 {
@@ -1515,54 +1695,3 @@ gboolean gok_spy_accessible_is_desktopChild(Accessible* accessible)
 	gok_log_leave();
 }
 */
-/* 
-* gok_spy_get_applications 
-* 
-* User must call gok_spy_free when finished with this list.
-* 
-* Returns: pointer to the list or NULL
-*/
-/*
-AccessibleNode* gok_spy_get_applications( )
-{
-	gint i;
-	gint j;
-	gint ndesktops;
-	gint nchildren;
-	Accessible* desktop;
-	Accessible* child;
-	AccessibleNode* plist;
-
-	gok_log_enter();
-
-	plist = NULL;
-	ndesktops = SPI_getDesktopCount();
-	for (i = 0; i < ndesktops; i++)
-	{
-		desktop = SPI_getDesktop(i);
-		gok_spy_accessible_implicit_ref(desktop);
-		if (desktop == NULL)
-		{
-			gok_log("desktop disappeared!?");
-			break;
-		}
-		nchildren = Accessible_getChildCount(desktop);
-		for (j = 0; j < nchildren; j++)
-		{
-			child = Accessible_getChildAtIndex(desktop, j);
-			gok_spy_accessible_implicit_ref(child);
-			if (child != NULL)
-			{
-				plist = gok_spy_append_node(plist, child);
-				if (plist == NULL)
-				{
-					gok_spy_accessible_unref(child);
-				}
-			}
-		}
-		gok_spy_accessible_unref(desktop);
-	}
-	gok_log_leave();
-	return plist;
-}
-*/
